@@ -1,40 +1,33 @@
 """Configuration Helpers for TurboGears 2"""
+
 import atexit
 import os
 import logging
+import warnings
 from copy import copy
 import mimetypes
 from UserDict import DictMixin
 
 from pylons.i18n import ugettext
-from genshi.filters import Translator
 
-from pylons import config as pylons_config
+from pylons.configuration import config as pylons_config
 from beaker.middleware import SessionMiddleware, CacheMiddleware
 from paste.cascade import Cascade
 from paste.registry import RegistryManager
 from paste.urlparser import StaticURLParser
-from paste.deploy.converters import asbool
-from pylons.middleware import report_libs, StatusCodeRedirect
+from paste.deploy.converters import asbool, asint
 
 import tg
 from tg import TGApp
-from tg.error import ErrorHandler
 from tg.util import Bunch, get_partial_dict, DottedFileNameFinder
 
 from routes import Mapper
 from routes.middleware import RoutesMiddleware
 from webob import Request
 
-from tw.api import make_middleware as tw_middleware
-
 log = logging.getLogger(__name__)
 
-class ConfigurationError(Exception):
-    """Exception raised for errors in the configuration."""
-
-    def __init__(self, message):
-        self.message = message
+class TGConfigError(Exception):pass
 
 class PylonsConfigWrapper(DictMixin):
     """Wrapper for the Pylons configuration.
@@ -101,7 +94,6 @@ class PylonsConfigWrapper(DictMixin):
 #Create a config object that has attribute style lookup built in.
 config = PylonsConfigWrapper(pylons_config)
 
-
 class AppConfig(Bunch):
     """Class to store application configuration.
 
@@ -112,7 +104,7 @@ class AppConfig(Bunch):
 
     AppConfig instances have a number of methods that are meant to be
     overridden by users who wish to have finer grained control over
-    the setup of the WSGI envirnment in which their application is run.
+    the setup of the WSGI environment in which their application is run.
 
     This is the place to configure custom routes, transaction handling,
     error handling, etc.
@@ -133,31 +125,52 @@ class AppConfig(Bunch):
         self.auto_reload_templates = True
         self.auth_backend = None
         self.default_renderer = 'genshi'
-        self.serve_static = True
         self.stand_alone = True
 
         # this is to activate the legacy renderers
         # legacy renderers are buffet interface plugins
         self.use_legacy_renderer = False
-        # if this is set to True the @expose decorator will be able to
-        # specify template names using a dotted name that will be searched
-        # in the python path. This option is used in tg.render.render_genshi
-        # TODO: we should set this to False once we implement simple names
-        # support in the @expose decorator as explained in #1942
-        # for the moment only the dotted names notation is supported with the
-        # new generation renderer functions
-        self.use_dotted_templatenames = True
 
+        self.use_ming = False
+        self.use_sqlalchemy = False
         self.use_toscawidgets = True
         self.use_transaction_manager = True
+        self.use_toscawidgets2 = False
 
-        #Registy for functions to be called on startup/teardown
+        # Registry for functions to be called on startup/teardown
         self.call_on_startup = []
         self.call_on_shutdown = []
+        self.controller_wrappers = []
+        self.hooks = dict(before_validate=[],
+                          before_call=[],
+                          before_render=[],
+                          after_render=[],
+                          before_config=[],
+                          after_config=[])
         # The codes TG should display an error page for. All other HTTP errors are
         # sent to the client or left for some middleware above us to handle
         self.handle_status_codes = [403, 404]
 
+        #override this variable to customize how the tw2 middleware is set up
+        self.custom_tw2_config = {}
+
+    def get_root_module(self):
+        root_module_path = self.paths['root']
+        base_controller_path = self.paths['controllers']
+        controller_path = base_controller_path[len(root_module_path)+1:]
+        root_controller_module = '.'.join([self.package.__name__] + controller_path.split(os.sep) + ['root'])
+        return root_controller_module
+
+    def register_hook(self, hook_name, func):
+        if hook_name == 'startup':
+            self.call_on_startup.append(func)
+        elif hook_name == 'shutdown':
+            self.call_on_shutdown.append(func)
+        elif hook_name == 'controller_wrapper':
+            self.controller_wrappers.append(func)
+        else:
+            self.hooks[hook_name].append(func)
+
     def setup_startup_and_shutdown(self):
         for cmd in self.call_on_startup:
             if callable(cmd):
@@ -189,24 +202,85 @@ class AppConfig(Bunch):
     def init_config(self, global_conf, app_conf):
         """Initialize the config object.
 
-        tg.config is a proxy for pylons.config that allows attribute style
-        access, so it's automatically setup when we create the pylons config.
+        tg.config is a proxy for pylons.configuration.config that allows
+        attribute style access, so it's automatically setup when we create
+        the pylons config.
 
         Besides basic initialization, this method copies all the values
-        in base_config into the ``pylons.config`` and ``tg.config`` objects.
+        in base_config into the ``pylons.configuration.config`` and
+        ``tg.config`` objects.
 
         """
         pylons_config.init_app(global_conf, app_conf,
                         package=self.package.__name__,
                         paths=self.paths)
+
+        self.auto_reload_templates = asbool(config.get('auto_reload_templates', True))
+        pylons_config['application_root_module'] = self.get_root_module()
+
         config.update(self)
+        # set up the response options to None.  This allows
+        # you to set the proper content type within a controller method
+        # if you choose.
+        pylons_config['pylons.response_options']['headers']['Content-Type'] = None
+
+        #see http://trac.turbogears.org/ticket/2247
+        if asbool(config['debug']):
+            config['pylons.strict_tmpl_context'] = True
+        else:
+            config['pylons.strict_tmpl_context'] = False
+
+        self.after_init_config()
+
+    def after_init_config(self):
+        """
+        Override this method to set up configuration variables at the application
+        level.  This method will be called after your configuration object has
+        been initialized on startup.  Here is how you would use it to override
+        the default setting of pylons.strict_tmpl_context ::
+
+            from tg.configuration import AppConfig
+            from pylons import config
+
+            class MyAppConfig(AppConfig):
+                def after_init_config(self):
+                    config['pylons.strict_tmpl_context'] = False
+
+            base_config = MyAppConfig()
+
+        """
 
     def setup_routes(self):
         """Setup the default TG2 routes
 
-        Overide this and setup your own routes maps if you want to use
+        Override this and setup your own routes maps if you want to use
         custom routes.
 
+        It is recommended that you keep the existing application routing in
+        tact, and just add new connections to the mapper above the routes_placeholder
+        connection.  Lets say you want to add a pylons controller SamplesController,
+        inside the controllers/samples.py file of your application.  You would
+        augment the app_cfg.py in the following way::
+
+            from routes import Mapper
+            from tg.configuration import AppConfig
+
+            class MyAppConfig(AppConfig):
+                def setup_routes(self):
+                    map = Mapper(directory=config['pylons.paths']['controllers'],
+                                always_scan=config['debug'])
+
+                    # Add a Samples route
+                    map.connect('/samples/', controller='samples', action=index)
+
+                    # Setup a default route for the root of object dispatch
+                    map.connect('*url', controller='root', action='routes_placeholder')
+
+                    config['routes.map'] = map
+
+
+            base_config = MyAppConfig()
+
         """
 
         map = Mapper(directory=config['pylons.paths']['controllers'],
@@ -228,21 +302,26 @@ class AppConfig(Bunch):
         config['pylons.app_globals'] = self.package.lib.app_globals.Globals()
         g = config['pylons.app_globals']
         g.dotted_filename_finder = DottedFileNameFinder()
-        config['pylons.helpers'] = self.package.lib.helpers
-        config['pylons.h'] = self.package.lib.helpers
 
     def setup_sa_auth_backend(self):
         """This method adds sa_auth information to the config."""
+
+        if 'beaker.session.secret' not in config:
+            raise TGConfigError("You must provide a value for 'beaker.session.secret'  If this is a project quickstarted with TG 2.0.2 or earlier \
+double check that you have base_config['beaker.session.secret'] = 'mysecretsecret' in your app_cfg.py file.")
+
         defaults = {
-                    'form_plugin': None
+                    'form_plugin': None,
+                    'cookie_secret': config['beaker.session.secret']
                    }
+
         # The developer must have defined a 'sa_auth' section, because
         # values such as the User, Group or Permission classes must be
         # explicitly defined.
         config['sa_auth'] = defaults
         config['sa_auth'].update(self.sa_auth)
 
-    def setup_mako_renderer(self):
+    def setup_mako_renderer(self, use_dotted_templatenames=None):
         """Setup a renderer and loader for mako templates.
 
         Override this to customize the way that the mako template
@@ -255,44 +334,59 @@ class AppConfig(Bunch):
         provided by tg.render.
 
         """
-        from tg.dottednamesupport import DottedTemplateLookup
-        from mako.lookup import TemplateLookup
 
         from tg.render import render_mako
 
+        if not use_dotted_templatenames:
+            use_dotted_templatenames = asbool(config.get('use_dotted_templatenames', 'true'))
 
-        if config.get('use_dotted_templatenames', False):
+
+        # If no dotted names support was required we will just setup
+        # a file system based template lookup mechanism.
+        compiled_dir = tg.config.get('templating.mako.compiled_templates_dir', None)
+
+        if not compiled_dir:
+            # Try each given templates path (when are they > 1 ?) for writability..
+            for template_path in self.paths['templates']:
+                if os.access(template_path, os.W_OK):
+                    compiled_dir = template_path
+                    break # first match is as good as any
+
+            # Last recourse: project-dir/data/templates (pylons' default directory)
+            if not compiled_dir:
+                try:
+                    root = os.path.dirname(os.path.abspath(self.package.__file__))
+                except AttributeError:
+                    # Thrown during unit tests when self.package.__file__ doesn't exist
+                    root = None
+
+                if root:
+                    pylons_default_path = os.path.join(root, '../data/templates')
+                    if os.access(pylons_default_path, os.W_OK):
+                        compiled_dir = pylons_default_path
+
+                if not compiled_dir:
+                    if use_dotted_templatenames:
+                        # Gracefully digress to in-memory template caching
+                        pass
+                    else:
+                        raise IOError("None of your templates directory, %s, are "
+                            "writable for compiled templates. Please set the "
+                            "templating.mako.compiled_templates_dir variable in your "
+                            ".ini file" % str(self.paths['templates']))
+
+        if use_dotted_templatenames:
             # Support dotted names by injecting a slightly different template
-            # lookup system that will return templates from dotted template
-            # notation.
+            # lookup system that will return templates from dotted template notation.
+            from tg.dottednames.mako_lookup import DottedTemplateLookup
             config['pylons.app_globals'].mako_lookup = DottedTemplateLookup(
                 input_encoding='utf-8', output_encoding='utf-8',
                 imports=['from webhelpers.html import escape'],
+                module_directory=compiled_dir,
                 default_filters=['escape'])
 
         else:
-            compiled_dir = tg.config.get('templating.mako.compiled_templates_dir', None)
-
-            if not compiled_dir:
-                # no specific compile dir give by conf... we expect that
-                # the server will have access to the first template dir
-                # to write the compiled version...
-                # If this is not the case we are doomed and the user should
-                # provide us the required config...
-                compiled_dir = self.paths['templates'][0]
-
-            # If no dotted names support was required we will just setup
-            # a file system based template lookup mechanism.
-            compiled_dir = tg.config.get('templating.mako.compiled_templates_dir', None)
-
-            if not compiled_dir:
-                # no specific compile dir give by conf... we expect that
-                # the server will have access to the first template dir
-                # to write the compiled version...
-                # If this is not the case we are doomed and the user should
-                # provide us the required config...
-                compiled_dir = self.paths['templates'][0]
-
+            from mako.lookup import TemplateLookup
             config['pylons.app_globals'].mako_lookup = TemplateLookup(
                 directories=self.paths['templates'],
                 module_directory=compiled_dir,
@@ -303,17 +397,19 @@ class AppConfig(Bunch):
 
         self.render_functions.mako = render_mako
 
-    def setup_chameleongenshi_renderer(self):
+    def setup_chameleon_genshi_renderer(self):
         """Setup a renderer and loader for the chameleon.genshi engine."""
-        from chameleon.genshi.loader import TemplateLoader as ChameleonLoader
-        from tg.render import render_chameleon_genshi
+        from tg.render import RenderChameleonGenshi
 
-        loader = ChameleonLoader(search_path=self.paths.templates,
+        if config.get('use_dotted_templatenames', True):
+            from tg.dottednames.chameleon_genshi_lookup \
+                import ChameleonGenshiTemplateLoader as TemplateLoader
+        else:
+            from chameleon.genshi.loader import TemplateLoader
+        loader = TemplateLoader(search_path=self.paths.templates,
                                 auto_reload=self.auto_reload_templates)
 
-        config['pylons.app_globals'].chameleon_genshi_loader = loader
-
-        self.render_functions.chameleon_genshi = render_chameleon_genshi
+        self.render_functions.chameleon_genshi = RenderChameleonGenshi(loader)
 
     def setup_genshi_renderer(self):
         """Setup a renderer and loader for Genshi templates.
@@ -322,40 +418,86 @@ class AppConfig(Bunch):
         filter, template loader
 
         """
-        from genshi.template import TemplateLoader
-        from tg.render import render_genshi
+        from tg.render import RenderGenshi
+        from genshi.filters import Translator
 
         def template_loaded(template):
             """Plug-in our i18n function to Genshi, once the template is loaded.
 
-            This function will be called by genshi TemplateLoader after
+            This function will be called by the Genshi TemplateLoader after
             loading the template.
 
             """
-            template.filters.insert(0, Translator(ugettext))
+            translator = Translator(ugettext)
+            template.filters.insert(0, translator)
+            if hasattr(template, 'add_directives'):
+                template.add_directives(Translator.NAMESPACE, translator)
 
+        if config.get('use_dotted_templatenames', True):
+            from tg.dottednames.genshi_lookup \
+                import GenshiTemplateLoader as TemplateLoader
+        else:
+            from genshi.template import TemplateLoader
         loader = TemplateLoader(search_path=self.paths.templates,
+                                max_cache_size=asint(self.get('genshi.max_cache_size', 30)),
                                 auto_reload=self.auto_reload_templates,
                                 callback=template_loaded)
 
-        config['pylons.app_globals'].genshi_loader = loader
+        self.render_functions.genshi = RenderGenshi(loader)
 
-        self.render_functions.genshi = render_genshi
+    def setup_kajiki_renderer(self):
+        """Setup a renderer and loader for the fastpt engine."""
+        from kajiki.loader import PackageLoader
+        from tg.render import render_kajiki
+        loader = PackageLoader()
+        config['pylons.app_globals'].kajiki_loader = loader
+        self.render_functions.kajiki = render_kajiki
 
     def setup_jinja_renderer(self):
         """Setup a renderer and loader for Jinja2 templates."""
         from jinja2 import ChoiceLoader, Environment, FileSystemLoader
+        from jinja2.filters import FILTERS
         from tg.render import render_jinja
 
+        if not 'jinja_extensions' in self :
+            self.jinja_extensions = []
+
+        if not 'jinja_filters' in self:
+            self.jinja_filters = {}
+
         config['pylons.app_globals'].jinja2_env = Environment(loader=ChoiceLoader(
-                [FileSystemLoader(path) for path in self.paths['templates']]),
-                auto_reload=self.auto_reload_templates)
+                 [FileSystemLoader(path) for path in self.paths['templates']]),
+                 auto_reload=self.auto_reload_templates, extensions=self.jinja_extensions)
+
+        # Try to load custom filters module under app_package.lib.templatetools
+        try:
+            filter_package = self.package.__name__ + ".lib.templatetools"
+            autoload_lib = __import__(filter_package, {}, {}, ['jinja_filters'])
+            autoload_filters = autoload_lib.jinja_filters.__dict__
+        except (ImportError, AttributeError):
+            autoload_filters = {}
+
+        # Add jinja filters
+        filters = dict(FILTERS, **autoload_filters)
+        filters.update(self.jinja_filters)
+        config['pylons.app_globals'].jinja2_env.filters = filters
 
         # Jinja's unable to request c's attributes without strict_c
+        warnings.simplefilter("ignore")
         config['pylons.strict_c'] = True
+        warnings.resetwarnings()
+        config['pylons.strict_tmpl_context'] = True
+
 
         self.render_functions.jinja = render_jinja
 
+    def setup_amf_renderer(self):
+        from tg.amfify import render_amf
+        self.render_functions.amf = render_amf
+
+    def setup_json_renderer(self):
+        from tg.render import render_json
+        self.render_functions.json = render_json
 
     def setup_default_renderer(self):
         """Setup template defaults in the buffed plugin.
@@ -368,6 +510,7 @@ class AppConfig(Bunch):
         config['buffet.template_engines'].pop()
         template_location = '%s.templates' % self.package.__name__
         template_location = '%s.templates' % self.package.__name__
+        from genshi.filters import Translator
 
         def template_loaded(template):
             template.filters.insert(0, Translator(ugettext))
@@ -390,14 +533,79 @@ class AppConfig(Bunch):
         for key, value in lookup.iteritems():
             mimetypes.add_type(value, key)
 
+    def setup_persistence(self):
+        """Override this method to define how your application configures it's persistence model.
+           the default is to setup sqlalchemy from the cofiguration file, but you might choose
+           to set up a persistence system other than sqlalchemy, or add an additional persistence
+           layer.  Here is how you would go about setting up a ming (mongo) persistence layer::
+
+            class MingAppConfig(AppConfig):
+                def setup_persistence(self):
+                    self.ming_ds = DataStore(config['mongo.url'])
+                    session = Session.by_name('main')
+                    session.bind = self.ming_ds
+        """
+        if self.use_sqlalchemy:
+            self.setup_sqlalchemy()
+        elif self.use_ming:
+            self.setup_ming()
+
+    def setup_ming(self):
+        """Setup MongoDB database engine using Ming"""
+        from ming.datastore import DataStore
+
+        datastore = DataStore(config['ming.url'], database=config['ming.db'])
+        config['pylons.app_globals'].ming_datastore = datastore
+        self.package.model.init_model(datastore)
+
     def setup_sqlalchemy(self):
-        """Setup SQLAlchemy database engine."""
+        """Setup SQLAlchemy database engine.
+
+        The most common reason for modifying this method is to add
+        multiple database support.  To do this you might modify your
+        app_cfg.py file in the following manner::
+
+            from tg.configuration import AppConfig, config
+            from pylons import config as pylons_config
+            from myapp.model import init_model
+
+            # add this before base_config =
+            class MultiDBAppConfig(AppConfig):
+                def setup_sqlalchemy(self):
+                    '''Setup SQLAlchemy database engine(s)'''
+                    from sqlalchemy import engine_from_config
+                    engine1 = engine_from_config(pylons_config, 'sqlalchemy.first.')
+                    engine2 = engine_from_config(pylons_config, 'sqlalchemy.second.')
+                    # engine1 should be assigned to sa_engine as well as your first engine's name
+                    config['pylons.app_globals'].sa_engine = engine1
+                    config['pylons.app_globals'].sa_engine_first = engine1
+                    config['pylons.app_globals'].sa_engine_second = engine2
+                    # Pass the engines to init_model, to be able to introspect tables
+                    init_model(engine1, engine2)
+
+            #base_config = AppConfig()
+            base_config = MultiDBAppConfig()
+
+        This will pull the config settings from your .ini files to create the necessary
+        engines for use within your application.  Make sure you have a look at :ref:`multidatabase`
+        for more information.
+
+        """
         from sqlalchemy import engine_from_config
         engine = engine_from_config(pylons_config, 'sqlalchemy.')
         config['pylons.app_globals'].sa_engine = engine
         # Pass the engine to initmodel, to be able to introspect tables
         self.package.model.init_model(engine)
 
+    def setup_auth(self):
+        """
+           Override this method to define how you would like the auth to be set up for your app.
+
+           For the standard TurboGears App, this will set up the auth with SQLAlchemy.
+        """
+        if self.auth_backend in ("ming", "sqlalchemy"):
+            self.setup_sa_auth_backend()
+
     def make_load_environment(self):
         """Return a load_environment function.
 
@@ -409,45 +617,45 @@ class AppConfig(Bunch):
         """
 
         def load_environment(global_conf, app_conf):
-            """Configure the Pylons environment via ``pylons.config``."""
+            """Configure the Pylons environment via ``pylons.configuration.config``."""
             global_conf=Bunch(global_conf)
             app_conf=Bunch(app_conf)
-            #Regesters functions to be called at startup and shutdown
-            #from self.call_on_startup and shutdown respectively.
-            self.setup_startup_and_shutdown()
 
             self.setup_paths()
             self.init_config(global_conf, app_conf)
+
+            #Registers functions to be called at startup and shutdown
+            #from self.call_on_startup and shutdown respectively.
+            self.setup_startup_and_shutdown()
+
             self.setup_routes()
             self.setup_helpers_and_globals()
             self.setup_mimetypes()
+            self.setup_auth()
 
-            if self.auth_backend == "sqlalchemy":
-                self.setup_sa_auth_backend()
+            if not 'json' in self.renderers: self.renderers.append('json')
 
-            if 'genshi' in self.renderers:
-                self.setup_genshi_renderer()
+            for renderer in self.renderers:
+                setup = getattr(self, 'setup_%s_renderer'%renderer, None)
+                if setup:
+                    setup()
+                else:
+                    raise Exception('This configuration object does not support the %s renderer'%renderer)
 
-            if 'chameleon_genshi' in self.renderers:
-                self.setup_chameleongenshi_renderer()
-
-            if 'mako' in self.renderers:
-                self.setup_mako_renderer()
-
-            if 'jinja' in self.renderers:
-                self.setup_jinja_renderer()
 
             if self.use_legacy_renderer:
                 self.setup_default_renderer()
 
-            if self.use_sqlalchemy:
-                self.setup_sqlalchemy()
+            self.setup_persistence()
 
         return load_environment
 
 
     def add_error_middleware(self, global_conf, app):
         """Add middleware which handles errors and exceptions."""
+        from pylons.middleware import report_libs, StatusCodeRedirect
+        from tg.error import ErrorHandler
+
         app = ErrorHandler(app, global_conf, **config['pylons.errorware'])
 
         # Display error documents for self.handle_status_codes status codes (and
@@ -462,16 +670,15 @@ class AppConfig(Bunch):
     def add_auth_middleware(self, app, skip_authentication):
         """
         Configure authentication and authorization.
-        
+
         :param app: The TG2 application.
         :param skip_authentication: Should authentication be skipped if
             explicitly requested? (used by repoze.who-testutil)
         :type skip_authentication: bool
-        
+
         """
-        from repoze.what.plugins.quickstart import setup_sql_auth
         from repoze.what.plugins.pylonshq import booleanize_predicates
-        
+
         # Predicates booleanized:
         booleanize_predicates()
 
@@ -481,33 +688,134 @@ class AppConfig(Bunch):
 
         # Removing keywords not used by repoze.who:
         auth_args = copy(self.sa_auth)
+        if 'sa_auth' in config:
+            auth_args.update(config.sa_auth)
         if 'password_encryption_method' in auth_args:
             del auth_args['password_encryption_method']
         if not skip_authentication:
             if not 'cookie_secret' in auth_args.keys():
-                msg = "base_config.sa_auth.cookie_secret is required " \
-                "you must define it in app_cfg.py or set " \
+                msg = "base_config.sa_auth.cookie_secret is required "\
+                "you must define it in app_cfg.py or set "\
                 "sa_auth.cookie_secret in development.ini"
-                print msg
-                raise ConfigurationError(message=msg)
-        app = setup_sql_auth(app, skip_authentication=skip_authentication,
-                             **auth_args)
+                raise TGConfigError(msg)
+
+        if self.auth_backend == "sqlalchemy":
+            from repoze.what.plugins.quickstart import setup_sql_auth
+            app = setup_sql_auth(app, skip_authentication=skip_authentication, **auth_args)
+        elif self.auth_backend == "ming":
+            from tgming import setup_ming_auth
+            app = setup_ming_auth(app, skip_authentication=skip_authentication, **auth_args)
         return app
 
     def add_core_middleware(self, app):
-        """Add support for routes dispatch, sessions, and caching."""
+        """Add support for routes dispatch, sessions, and caching.
+        This is where you would want to override if you wanted to provide your
+        own routing, session, or caching middleware.  Your app_cfg.py might look something
+        like this::
+
+            from tg.configuration import AppConfig
+            from routes.middleware import RoutesMiddleware
+            from beaker.middleware import CacheMiddleware
+            from mysessionier.middleware import SessionMiddleware
+
+            class MyAppConfig(AppConfig):
+                def add_core_middleware(self, app):
+                    app = RoutesMiddleware(app, config['routes.map'])
+                    app = SessionMiddleware(app, config)
+                    app = CacheMiddleware(app, config)
+                    return app
+            base_config = MyAppConfig()
+        """
         app = RoutesMiddleware(app, config['routes.map'])
         app = SessionMiddleware(app, config)
         app = CacheMiddleware(app, config)
         return app
 
     def add_tosca_middleware(self, app):
-        """Configure the ToscaWidgets middleware."""
-        app = tw_middleware(app, {
-            'toscawidgets.framework.default_view': self.default_renderer,
-            'toscawidgets.framework.translator': ugettext,
-            'toscawidgets.middleware.inject_resources': True,
-            })
+        """Configure the ToscaWidgets middleware.
+
+        If you would like to override the way the TW middleware works, you might do something like::
+
+            from tg.configuration import AppConfig
+            from tw.api import make_middleware as tw_middleware
+
+            class MyAppConfig(AppConfig):
+
+                def add_tosca2_middleware(self, app):
+
+                    app = tw_middleware(app, {
+                        'toscawidgets.framework.default_view': self.default_renderer,
+                        'toscawidgets.framework.translator': ugettext,
+                        'toscawidgets.middleware.inject_resources': False,
+                        })
+                    return app
+
+            base_config = MyAppConfig()
+
+
+
+        The above example would disable resource injection.
+
+        There is more information about the settings you can change
+        in the ToscaWidgets `middleware. <http://toscawidgets.org/documentation/ToscaWidgets/modules/middleware.html>`
+
+
+        """
+
+        from tw.api import make_middleware as tw_middleware
+
+
+        twconfig = {'toscawidgets.framework.default_view': self.default_renderer,
+                    'toscawidgets.framework.translator': ugettext,
+                    'toscawidgets.middleware.inject_resources': True,
+                    }
+        for k,v in config.iteritems():
+            if k.startswith('toscawidgets.framework.') or k.startswith('toscawidgets.middleware.'):
+                twconfig[k] = v
+
+        if 'toscawidgets.framework.resource_variant' in config:
+            import tw.api
+            tw.api.resources.registry.ACTIVE_VARIANT = config['toscawidgets.framework.resource_variant']
+            #remove it from the middleware madness
+            del twconfig['toscawidgets.framework.resource_variant']
+
+        app = tw_middleware(app, twconfig)
+        return app
+
+    def add_tosca2_middleware(self, app):
+        """Configure the ToscaWidgets2 middleware.
+
+        If you would like to override the way the TW2 middleware works,
+        you might do change your app_cfg.py to add something like::
+
+            from tg.configuration import AppConfig
+            from tw2.core.middleware import TwMiddleware
+
+            class MyAppConfig(AppConfig):
+
+                def add_tosca2_middleware(self, app):
+
+                    app = TwMiddleware(app,
+                        default_engine=self.default_renderer,
+                        translator=ugettext,
+                        auto_reload_templates = False
+                        )
+
+                    return app
+            base_config = MyAppConfig()
+
+
+
+        The above example would always set the template auto reloading off. (This is normally an
+        option that is set within your application's ini file.)
+        """
+        from tw2.core.middleware import Config, TwMiddleware
+        default_tw2_config = dict( default_engine=self.default_renderer,
+                                   translator=ugettext,
+                                   auto_reload_templates=asbool(self.get('templating.mako.reloadfromdisk', 'false'))
+                                   )
+        default_tw2_config.update(self.custom_tw2_config)
+        app = TwMiddleware(app, **default_tw2_config)
         return app
 
     def add_static_file_middleware(self, app):
@@ -543,6 +851,11 @@ class AppConfig(Bunch):
         from repoze.tm import make_tm
         return make_tm(app, self.commit_veto)
 
+    def add_ming_middleware(self, app):
+        """Set up the ming middleware for the unit of work"""
+        import ming.orm.middleware
+        return ming.orm.middleware.MingMiddleware(app)
+
     def add_dbsession_remover_middleware(self, app):
         """Set up middleware that cleans up the sqlalchemy session.
 
@@ -601,13 +914,23 @@ class AppConfig(Bunch):
             app = TGApp()
             if wrap_app:
                 app = wrap_app(app)
+
+            for hook in self.hooks['before_config']:
+                app = hook(app)
+
+            avoid_sess_touch = config.get('beaker.session.tg_avoid_touch', 'false')
+            config['beaker.session.tg_avoid_touch'] = asbool(avoid_sess_touch)
+
             app = self.add_core_middleware(app)
 
             if self.use_toscawidgets:
                 app = self.add_tosca_middleware(app)
 
-            if self.auth_backend == "sqlalchemy":
-                # Skipping authentication if explicitly requested. Used by 
+            if self.use_toscawidgets2:
+                app = self.add_tosca2_middleware(app)
+
+            if self.auth_backend:
+                # Skipping authentication if explicitly requested. Used by
                 # repoze.who-testutil:
                 skip_authentication = app_conf.get('skip_authentication', False)
                 app = self.add_auth_middleware(app, skip_authentication)
@@ -622,8 +945,16 @@ class AppConfig(Bunch):
                     self.DBSession = self.model.DBSession
                 app = self.add_dbsession_remover_middleware(app)
 
-            app = maybe_make_body_seekable(app)
+            if self.use_ming:
+                app = self.add_ming_middleware(app)
 
+            if pylons_config.get('make_body_seekable'):
+                app = maybe_make_body_seekable(app)
+
+            if 'PYTHONOPTIMIZE' in os.environ:
+                warnings.warn("Forcing full_stack=False due to PYTHONOPTIMIZE enabled. "+\
+                              "Error Middleware will be disabled", RuntimeWarning, stacklevel=2)
+                full_stack = False
 
             if asbool(full_stack):
                 if (self.auth_backend is None
@@ -632,7 +963,7 @@ class AppConfig(Bunch):
                     # responses we redirect those responses to a nicely
                     # formatted error page
                     self.handle_status_codes.append(401)
-                # This should nevery be true for internal nested apps
+                # This should never be true for internal nested apps
                 app = self.add_error_middleware(global_conf, app)
 
             # Establish the registry for this application
@@ -640,17 +971,23 @@ class AppConfig(Bunch):
 
             # Static files (if running in production, and Apache or another
             # web server is serving static files)
+
+            #if the user has set the value in app_config, don't pull it from the ini
+            if not hasattr(self, 'serve_static'):
+                self.serve_static = asbool(config.get('serve_static', 'true'))
             if self.serve_static:
                 app = self.add_static_file_middleware(app)
 
+            for hook in self.hooks['after_config']:
+                app = hook(app)
+
             return app
 
         return make_base_app
 
 def maybe_make_body_seekable(app):
     def wrapper(environ, start_response):
-        if pylons_config.get('make_body_seekable'):
-            log.debug("Making request body seekable")
-            Request(environ).make_body_seekable()
+        log.debug("Making request body seekable")
+        Request(environ).make_body_seekable()
         return app(environ, start_response)
     return wrapper
