@@ -8,40 +8,46 @@ the functions they wrap, and then the DecoratedController provides the hooks
 needed to support these decorators.
 
 """
-import inspect
+from warnings import warn
 import formencode
 from paste.util.mimeparse import best_match
 from decorator import decorator
 
-#this can be removed after tg_format is removed
-import mimetypes
-import warnings
-
-from webob.exc import HTTPUnauthorized
+from webob.exc import HTTPUnauthorized, HTTPMethodNotAllowed
 from webob.multidict import MultiDict
-from webhelpers.paginate import Page
-from pylons import config, request, response
+from tg.paginate import Page
+from pylons.configuration import config
+from pylons import request, response
 from pylons.controllers.util import abort
-from pylons import tmpl_context as c
+from formencode import variabledecode
+from tg import tmpl_context
 from tg.util import partial
-from repoze.what.plugins.pylonshq import ActionProtector, ControllerProtector
+
+try:
+    from repoze.what.plugins.pylonshq import ActionProtector
+    from repoze.what.plugins.pylonshq.protectors import _BaseProtectionDecorator
+except ImportError:
+    class ActionProtector(object):
+        pass
+    class _BaseProtectionDecorator(object):
+        pass
 
 from tg.configuration import Bunch
 from tg.flash import flash
 #from tg.controllers import redirect
 
+
 class Decoration(object):
     """ Simple class to support 'simple registration' type decorators
     """
     def __init__(self, controller):
         self.controller = controller
         self.engines = {}
+        self.engines_keys = []
         self.custom_engines = {}
         self.render_custom_format = None
         self.validation = None
         self.error_handler = None
-        self.is_default_controller = False
-        self.is_lookup_controller = False
         self.hooks = dict(before_validate=[],
                           before_call=[],
                           before_render=[],
@@ -53,14 +59,22 @@ class Decoration(object):
         return func.decoration
     get_decoration = classmethod(get_decoration)
 
+    @property
     def exposed(self):
-        return bool(self.engines)
-    exposed = property(exposed)
+        return bool(self.engines) or bool(self.custom_engines)
 
     def run_hooks(self, hook, *l, **kw):
+        for func in config.get('hooks', {}).get(hook, []):
+            func(*l, **kw)
         for func in self.hooks[hook]:
             func(*l, **kw)
 
+    def wrap_controller(self, controller):
+        controller_callable = controller
+        for wrapper in config.get('controller_wrappers', []):
+            controller_callable = wrapper(self, controller_callable)
+        return controller_callable
+
     def register_template_engine(self, content_type, engine, template,
                                  exclude_names):
         """Registers an engine on the controller.
@@ -76,16 +90,20 @@ class Decoration(object):
         from JSONification, and other 'automatic' engines which don't
         require a template.
         """
-
         if content_type is None:
             content_type = '*/*'
         self.engines[content_type] = engine, template, exclude_names
 
+        #this is a work-around to make text/html prominent in respect
+        #to other common choices when they have the same weight for
+        # paste.util.mimeparse.best_match.
+        self.engines_keys = sorted(self.engines.keys(), reverse=True)
+
     def register_custom_template_engine(self, custom_format, content_type, engine, template,
                                         exclude_names):
         """Registers a custom engine on the controller.
 
-        Mulitple engines can be registered, but only one engine per
+        Multiple engines can be registered, but only one engine per
         custom_format.
 
         The engine is registered when @expose is used with the
@@ -111,37 +129,46 @@ class Decoration(object):
         content_type, template, and exclude_names for a particular
         tg_format (which is pulled off of the request headers).
         """
-        #remove this after deprecation period for tg_format
-        tg_format = request.headers.get('tg_format')
 
         if hasattr(request, 'response_type') and request.response_type in self.engines:
             accept_types = request.response_type
-
-        elif tg_format:
-            warnings.warn('tg_format is now deprecated.  Use .mimetype in your URL to create the same behavior')
-            if '/' not in tg_format:
-                accept_types = mimetypes.guess_type('.'+tg_format)[0]
-                if accept_types is None:
-                    raise Exception('Unknown mimetype: %s'%tg_format)
-            else:
-                accept_types = tg_format
         else:
             accept_types = request.headers.get('accept', '*/*')
 
-        if self.render_custom_format:
-            content_type, engine, template, exclude_names = self.custom_engines[self.render_custom_format]
+        try:
+            render_custom_format = request._render_custom_format[self.controller]
+        except:
+            render_custom_format = self.render_custom_format
 
+        if render_custom_format:
+            content_type, engine, template, exclude_names = self.custom_engines[render_custom_format]
         else:
-            content_type = best_match(self.engines.keys(), accept_types)
+            if self.engines:
+                content_type = best_match(self.engines_keys, accept_types)
+            else:
+                content_type = 'text/html'
+
+            if content_type == 'CUSTOM/LEAVE':
+                warn('@expose(CUSTOM_CONTENT_TYPE) is no longer needed and should be replaced with @expose()')
+
+            # check for overridden content type from the controller call
+            controller_content_type = response.headers.get('Content-Type')
+
+            if controller_content_type:
+                # make sure we handle content_types like 'text/html; charset=utf-8'
+                content_type = controller_content_type.split(';')[0]
+
             # check for overridden templates
             try:
-                engine, template, exclude_names = request._override_mapping[self.controller][content_type]
+                engine, template, exclude_names = request._override_mapping[self.controller][content_type.split(";")[0]]
             except (AttributeError, KeyError):
-                engine, template, exclude_names = self.engines[content_type]
+                engine, template, exclude_names = self.engines.get(content_type, (None, None, None))
 
 
-        if 'charset' not in content_type and (
-           content_type.startswith('text') or content_type  == 'application/json'):
+        if 'charset' not in content_type and (content_type.startswith('text') or \
+                                              content_type in ('application/xhtml+xml',
+                                                               'application/xml',
+                                                               'application/json')):
             content_type = '%s; charset=utf-8' % content_type
 
         return content_type, engine, template, exclude_names
@@ -262,8 +289,8 @@ class expose(object):
         if exclude_names is None:
             exclude_names = []
 
-        if template == 'json':
-            engine, template = 'json', ''
+        if template in config.get('renderers', []):
+            engine, template = template, ''
 
         elif ':' in template:
             engine, template = template.split(':', 1)
@@ -285,7 +312,7 @@ class expose(object):
             else:
                 content_type = 'text/html'
 
-        if engine == 'json' and 'tmpl_context' not in exclude_names:
+        if (engine == 'json' or engine == 'amf') and 'tmpl_context' not in exclude_names:
             exclude_names.append('tmpl_context')
 
         self.engine = engine
@@ -296,7 +323,6 @@ class expose(object):
 
     def __call__(self, func):
         deco = Decoration.get_decoration(func)
-
         if self.custom_format:
             deco.register_custom_template_engine(
                 self.custom_format, self.content_type, self.engine,
@@ -306,6 +332,7 @@ class expose(object):
                 self.content_type, self.engine, self.template, self.exclude_names)
         return func
 
+
 def use_custom_format(controller, custom_format):
     """Use use_custom_format in a controller in order to change
     the active @expose decorator when available."""
@@ -315,7 +342,11 @@ def use_custom_format(controller, custom_format):
     if not custom_format in deco.custom_engines.keys():
         raise ValueError("'%s' is not a valid custom_format" % custom_format)
 
-    deco.render_custom_format = custom_format
+    try:
+        render_custom_format = request._render_custom_format
+    except AttributeError:
+        render_custom_format = request._render_custom_format = {}
+    render_custom_format[controller.im_func] = custom_format
 
 def override_template(controller, template):
     """Use overide_template in a controller in order to change the
@@ -332,24 +363,19 @@ def override_template(controller, template):
     future versions may make the `genshi:` optional if you want to use
     the default engine.
     """
-    if hasattr(controller, 'decoration'):
-        decoration = controller.decoration
-    else:
-        return
-    if hasattr(decoration, 'engines'):
-        engines = decoration.engines
-    else:
-        return
-
-    text_engine = engines.get('text/html')
-    template = template.split(':')
-    template.extend(text_engine[2:])
     try:
-        override_mapping = request._override_mapping
-    except AttributeError:
-        override_mapping = request._override_mapping = {}
-    override_mapping[controller.im_func] = {"text/html" : template}
-
+        engines = controller.decoration.engines
+    except:
+        return
+    
+    for content_type, content_engine in engines.iteritems():
+        tmpl = template.split(':', 1)
+        tmpl.extend(content_engine[2:])
+        try:
+            override_mapping = request._override_mapping
+        except AttributeError:
+            override_mapping = request._override_mapping = {}
+        override_mapping.setdefault(controller.im_func, {}).update({content_type: tmpl})
 
 class validate(object):
     """Regesters which validators ought to be applied
@@ -385,21 +411,9 @@ class validate(object):
         deco.validation = self
         return func
 
-def default(func):
-    '''Registers a method as the "default" controller method'''
-    deco = Decoration.get_decoration(func)
-    deco.is_default_controller = True
-    return func
-        
-def lookup(func):
-    '''Registers a method as the "lookup" controller method'''
-    deco = Decoration.get_decoration(func)
-    deco.is_lookup_controller = True
-    return func
 
-def paginate(name, items_per_page=10, use_prefix=False):
-    """
-    Paginate a given collection.
+class paginate(object):
+    """Paginate a given collection.
 
     This decorator is mainly exposing the functionality
     of :func:`webhelpers.paginate`.
@@ -418,9 +432,7 @@ def paginate(name, items_per_page=10, use_prefix=False):
 
     To render the actual pager, use::
 
-      ${c.paginators.<name>.pager()}
-
-    where c is the tmpl_context.
+      ${tmpl_context.paginators.<name>.pager()}
 
     It is possible to have several :func:`paginate`-decorators for
     one controller action to paginate several collections independently
@@ -432,68 +444,72 @@ def paginate(name, items_per_page=10, use_prefix=False):
         the collection to be paginated.
       items_per_page
         the number of items to be rendered. Defaults to 10
+      max_items_per_page
+        the maximum number of items allowed to be set via parameter.
+        Defaults to 0 (does not allow to change that value).
       use_prefix
         if True, the parameters the paginate
         decorator renders and reacts to are prefixed with
         "<name>_". This allows for multi-pagination.
 
     """
-    prefix = ""
-    if use_prefix:
-        prefix = name + "_"
-    own_parameters = dict(
-        page="%spage" % prefix,
-        items_per_page="%sitems_per_page" % prefix
-        )
-    #@decorator
-    def _d(f):
-        def _w(*args, **kwargs):
-            page = int(kwargs.pop(own_parameters["page"], 1))
-            real_items_per_page = int(
-                    kwargs.pop(
-                            own_parameters['items_per_page'],
-                            items_per_page))
-            argvars = inspect.getargspec(f)[0][1:]
-            if argvars:
-                args = list(args)
-                for i, var in enumerate(args):
-                    if i>=len(argvars):
-                        break;
-                    var = argvars[i]
-                    if var in kwargs:
-                        args[i] = kwargs[var]
-                        del kwargs[var]
-            
-            res = f(*args, **kwargs)
-            if isinstance(res, dict) and name in res:
-                additional_parameters = MultiDict()
-                for key, value in request.str_params.iteritems():
-                    if key not in own_parameters:
-                        additional_parameters.add(key, value)
 
-                collection = res[name]
-                page = Page(
-                    collection,
-                    page,
-                    items_per_page=real_items_per_page,
-                    **additional_parameters.dict_of_lists()
-                    )
-                # wrap the pager so that it will render
-                # the proper page-parameter
-                page.pager = partial(page.pager,
-                        page_param=own_parameters["page"])
-                res[name] = page
-                # this is a bit strange - it appears
-                # as if c returns an empty
-                # string for everything it dosen't know.
-                # I didn't find that documented, so I
-                # just put this in here and hope it works.
-                if not hasattr(c, 'paginators') or type(c.paginators) == str:
-                    c.paginators = Bunch()
-                c.paginators[name] = page
-            return res
-        return _w
-    return _d
+    def __init__(self, name, use_prefix=False,
+        items_per_page=10, max_items_per_page=0):
+        self.name = name
+        prefix = use_prefix and name + '_' or ''
+        self.page_param = prefix + 'page'
+        self.items_per_page_param = prefix + 'items_per_page'
+        self.items_per_page = items_per_page
+        self.max_items_per_page = max_items_per_page
+
+    def __call__(self, func):
+        decoration = Decoration.get_decoration(func)
+        decoration.register_hook('before_validate', self.before_validate)
+        decoration.register_hook('before_render', self.before_render)
+        return func
+
+    def before_validate(self, remainder, params):
+        page = params.pop(self.page_param, None)
+        if page:
+            try:
+                page = int(page)
+                if page < 1:
+                    raise ValueError
+            except ValueError:
+                page = 1
+        else:
+            page = 1
+        request.paginate_page = page or 1
+        items_per_page = params.pop(self.items_per_page_param, None)
+        if items_per_page:
+            try:
+                items_per_page = min(
+                    int(items_per_page), self.max_items_per_page)
+                if items_per_page < 1:
+                    raise ValueError
+            except ValueError:
+                items_per_page = self.items_per_page
+        else:
+            items_per_page = self.items_per_page
+        request.paginate_items_per_page = items_per_page
+        request.paginate_params = params.copy()
+        if items_per_page != self.items_per_page:
+            request.paginate_params[self.items_per_page_param] = items_per_page
+
+    def before_render(self, remainder, params, output):
+        if not isinstance(output, dict) or not self.name in output:
+            return
+        collection = output[self.name]
+        page = Page(collection, request.paginate_page,
+            request.paginate_items_per_page, controller='/')
+        page.kwargs = request.paginate_params
+        if self.page_param != 'name':
+            page.pager = partial(page.pager, page_param=self.page_param)
+        if not getattr(tmpl_context, 'paginators', None):
+            tmpl_context.paginators = Bunch()
+        tmpl_context.paginators[self.name] = output[self.name] = page
+
 
 @decorator
 def postpone_commits(func, *args, **kwargs):
@@ -511,8 +527,33 @@ def postpone_commits(func, *args, **kwargs):
     s.commit = old_commit
     return retval
 
-@decorator
-def without_trailing_slash(func, *args, **kwargs):
+@before_validate
+def https(remainder, params):
+    '''Ensure that the decorated method is always called with https://'''
+    from tg.controllers import redirect
+    if request.scheme.lower() == 'https': return
+    if request.method.upper() == 'GET':
+        redirect('https' + request.url[len(request.scheme):])
+    raise HTTPMethodNotAllowed(headers=dict(Allow='GET')).exception
+
+@before_validate
+def variable_decode(remainder, params):
+    '''Best-effort formencode.variabledecode on the params before validation
+
+    If any exceptions are raised due to invalid parameter names, they are
+    silently ignored, hopefully to be caught by the actual validator.  Note that
+    this decorator will *add* parameters to the method, not remove.  So for
+    instnace a method will move from {'foo-1':'1', 'foo-2':'2'} to
+    {'foo-1':'1', 'foo-2':'2', 'foo':['1', '2']}.
+    '''
+    try:
+        new_params = variabledecode.variable_decode(params)
+        params.update(new_params)
+    except:
+        pass
+
+@before_validate
+def without_trailing_slash(remainder, params):
     """This decorator allows you to ensure that the URL does not end in "/"
     The decorator accomplish this by redirecting to the correct URL.
 
@@ -533,10 +574,10 @@ def without_trailing_slash(func, *args, **kwargs):
     if request.method == 'GET' and request.path.endswith('/') and not(request.response_type) and len(request.params)==0:
         from tg.controllers import redirect
         redirect(request.url[:-1])
-    return func(*args, **kwargs)
 
-@decorator
-def with_trailing_slash(func, *args, **kwargs):
+
+@before_validate
+def with_trailing_slash(remainder, params):
     """This decorator allows you to ensure that the URL ends in "/"
     The decorator accomplish this by redirecting to the correct URL.
 
@@ -554,10 +595,13 @@ def with_trailing_slash(func, *args, **kwargs):
     In the above example http://localhost:8080/sample redirects to http://localhost:8080/sample/
     In addition, the URL http://localhost:8080/sample/1 redirects to http://localhost:8080/sample/1/
     """
-    if request.method == 'GET' and not(request.path.endswith('/')) and not(request.response_type) and len(request.params)==0:
+    if (request.method == 'GET'
+        and not(request.path.endswith('/'))
+        and not(request.response_type)
+        and len(request.params)==0):
         from tg.controllers import redirect
         redirect(request.url+'/')
-    return func(*args, **kwargs)
+
 
 #{ Authorization decorators
 
@@ -565,15 +609,15 @@ def with_trailing_slash(func, *args, **kwargs):
 class require(ActionProtector):
     """
     TurboGears-specific repoze.what-pylons action protector.
-    
+
     The default authorization denial handler of this protector will flash
     the message of the unmet predicate with ``warning`` or ``error`` as the
     flash status if the HTTP status code is 401 or 403, respectively.
-    
+
     See :class:`allow_only` for controller-wide authorization.
-    
+
     """
-    
+
     def default_denial_handler(self, reason):
         """Authorization denial handler for repoze.what-pylons protectors."""
         if response.status_int == 401:
@@ -585,26 +629,46 @@ class require(ActionProtector):
         abort(response.status_int, reason)
 
 
-class allow_only(ControllerProtector):
+class allow_only(_BaseProtectionDecorator):
     """
     TurboGears-specific repoze.what-pylons controller protector.
-    
+
     The default authorization denial handler of this protector will flash
     the message of the unmet predicate with ``warning`` or ``error`` as the
     flash status if the HTTP status code is 401 or 403, respectively, since
     by default the ``__before__`` method of the controller is decorated with
     :class:`require`.
-    
+
     If the controller class has the ``_failed_authorization`` *class method*,
     it will replace the default denial handler.
-    
+
     """
     protector = require
-    
+
     def __call__(self, cls, *args, **kwargs):
+        if hasattr(self.protector, 'predicate'):
+            cls.allow_only=self.protector.predicate
         if hasattr(cls, '_failed_authorization'):
             self.denial_handler = cls._failed_authorization
-        return super(allow_only, self).__call__(cls, *args, **kwargs)
+        sup = super(allow_only, self)
+        if hasattr(sup, '__call__'):
+            return super(allow_only, self).__call__(cls, *args, **kwargs)
 
+class cached_property(object):
+    def __init__(self, func):
+        self.__name__ = func.__name__
+        self.__module__ = func.__module__
+        self.__doc__ = func.__doc__
+        self.func = func
+
+    def __get__(self, obj, type=None):
+        if obj is None:
+            return self
+
+        try:
+            value = obj.__dict__[self.__name__]
+        except KeyError:
+            value = obj.__dict__[self.__name__] = self.func(obj)
+        return value
 
 #}
